import numpy as np
from random import choice, uniform
from hula.rlutils import *

class ANN:
  def __init__(self, design):
    
    self.WS = []

    for dim in lagzip(design):
      self.WS.append(np.random.randn(*dim))

    self.design = design

    self.epoch = 0
    
    self.error = 1

  def _getWeights(self):
    return tuple([tuple([tuple(j) for j in i]) for i in self.WS])

  def _getAction(self):
    return tuple([tuple([tuple(j) for j in i]) for i in self.action])

  def activate(self, X):
    latest = np.array(X)
    for n, W in enumerate(self.WS):
      latest = latest.dot(W)
      if W.all() == self.WS[-1].all():
        latest = latest/5
      elif W.all() == self.WS[0].all():
        latest = latest/5
      else:
        latest = tanh(latest)*2
    return latest

  def __init__TrainDR(self):
    self.limboActions = []
    self.StateActions = {}
    self.action = []
    for dim in lagzip(self.design):
      self.action.append(np.random.randn(*dim)/100)
    self.epoch = 1

  def _actAddition(self):
    for n, a, w in zip(range(len(self.WS)), self.action, self.WS):
      self.WS[n] = w + a

  def _randomAction(self, alpha):
    self.action = []
    for dim in lagzip(self.design):
      self.action.append(np.random.randn(*dim)*alpha)
    
  def GenDR(self, X, Y, alpha):
    if self.epoch == 0:
      self.__init__TrainDR()

    ogerror = float(np.mean((self.activate(X) - Y)**2))

    currentW = self._getWeights()

    self.error = ogerror
    self._randomAction(alpha)
    self._actAddition()

    newerror = float(np.mean((self.activate(X) - Y)**2))

    self.StateActions[currentW] = {
      'action':self._getAction(),
      'score':(newerror - ogerror) / newerror
    }

  def _error(self, a, b):
    return TwDimDistance(a, b)

  def simplify(self, thresh, maxscore):
    newStateActions = {}
    finalStateActions = {}

    self.StateActions = {s:self.StateActions[s] for s in sorted(self.StateActions, key=lambda s: ThrDimMagnitude(s))}

    last_checkpoint = 0
    for x in self.StateActions:
      if ThrDimMagnitude(x) - last_checkpoint > thresh:
        newStateActions[x] = self.StateActions[x]
        last_checkpoint = ThrDimMagnitude(x)

    for state in newStateActions:
      if newStateActions[state]['score'] < maxscore:
        finalStateActions[state] = newStateActions[state]
    
    self.StateActions = finalStateActions

  def nonSpecificGenDR(self, alpha):
    if self.epoch == 0:
      self.__init__TrainDR()

    self._randomAction(alpha)

    self.limboActions.append([self._getWeights(), self._getAction()])

    self._actAddition()

  def nonSpecificScoreDR(self, score):

    for action in self.limboActions:
      self.StateActions[action[0]] = {
        'action':action[1],
        'score':score
      }

    self.limboActions = []

  def TrainDR(self, X, Y, alpha, genaction=True):
    
    if genaction:
      self.GenDR(X, Y, alpha)

    else:
      self.error = float(np.mean((self.activate(X) - Y)**2))
    
    self.epoch += 1

    currentState = self._getWeights()

    choices = []

    if self.epoch > 5:
      mnd = min([ThrDimDistance(currentState, state) for state in self.StateActions])
      mnscore = min([self.StateActions[state]['score'] for state in self.StateActions])

      mxd = max([ThrDimDistance(currentState, state)-mnd for state in self.StateActions])
      mxscore = max([self.StateActions[state]['score']-mnscore for state in self.StateActions])

      for state in self.StateActions:
        d, a = [(ThrDimDistance(currentState, state)-mnd)/mxd, (self.StateActions[state]['score']-mnscore)/mxscore]
        choices.append([state, (d*3 + a*7)/10, self.StateActions[state]['action']])

      for n, W, act in zip(range(len(self.WS)), self.WS, sorted(choices, key=lambda x: x[1])[0][2]):
        self.WS[n] = self.WS[n] + np.array(act) * (1 + alpha)

      self.error = float(np.mean((self.activate(X) - Y)**2))

  def split(self, layern):
    a1 = ANN(self.design[:layern])
    a2 = ANN(self.design[layern:])
    a1.WS = self.WS[:layern]
    a2.WS = self.WS[layern:]

    return (a1, a2)

  def nonSpecificTrainDR(self, alpha):

    self.epoch += 1

    currentState = self._getWeights()

    choices = []

    if self.epoch > 1:
      mxd = max([ThrDimDistance(currentState, state) for state in self.StateActions])
      mxscore = max([self.StateActions[state]['score'] for state in self.StateActions])

      for state in self.StateActions:
        d, a = [(ThrDimDistance(currentState, state))/mxd, (self.StateActions[state]['score'])/mxscore]
        choices.append([state, (d*3 + a*7)/10, self.StateActions[state]['action']])

      for n, W, act in zip(range(len(self.WS)), self.WS, sorted(choices, key=lambda x: x[1])[0][2]):
        self.WS[n] = self.WS[n] + np.array(act) * (1 + alpha)



from hula.MemoryF import *
from hula.RecursiveL import *
from hula.rlutils import *

class CNET:
  
  def __init__(self, design):
    self.mems = []
    
    for n in design[1:]:
      self.mems.append([MemoryActivation(0, [0, 0, 0, 0, 0, 0]) for _ in range(n)])

  def activate(self, X):
    
    latest = X

    for layer in self.mems:
      temp = []
      for mem in layer:
        temp.append(mem.multiactivate(latest))

      latest = temp

    return latest

  def act(self, alpha):
    for layer in self.mems:
      for mem in layer:
        mem.randomAct(alpha)

  def score(self, score):
    for layer in self.mems:
      for mem in layer:
        mem.score(score)

  def train(self, alpha):
    for layer in self.mems:
      for mem in layer:
        mem.train(alpha)

  def trim(self, perc):
    for layer in self.mems:
      for mem in layer:
        mem.trim(perc)
      





from math import e
import numpy as np
from random import uniform

def softmax(*args):
  return np.exp(args)/np.sum(np.exp(args))

def sigmoid(x):
  return 1/(1+np.exp(-x))

def tanh(x):
  return (np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-x))

def distance(x, y):
  return np.sum([(a-b)**2 for a, b in zip(x, y)])**(0.5)

def magnitude(x):
  return np.sum([a**2 for a in x])**(0.5)

class MemoryActivation:

  def __init__(self, starting_state, params):

    self.state = starting_state
    self.rw, self.uw, self.nw, self.rb, self.ub, self.nb = params

    self.stateActions = []
    self.limboActions = []


  def activate(self, x):

    replace, use = sigmoid(self.rw * x + self.rb), sigmoid(self.uw * x + self.ub)

    self.state = self.state * (1-replace) + x * replace
    
    return tanh(x * (1 - use) + self.state * use + self.nb)

  def multiactivate(self, X):
    o = []
    for x in X:
      o.append(self.activate(x))

    return sigmoid(sum(o))

  def setParams(self, params):
    self.rw, self.uw, self.nw, self.rb, self.ub, self.nb = params


  def getParams(self):
    return [self.rw, self.uw, self.nw, self.rb, self.ub, self.nb]


  def randomAct(self, alpha=0.01):

    action = [uniform(-alpha, alpha) for _ in range(6)]

    currentstate = self.getParams()

    self.setParams(np.add(self.getParams(), action))

    self.limboActions.append([currentstate, action])


  def score(self, score):

    for action in self.limboActions:
      self.stateActions.append({
        'action':action[1],
        'score':score,
        'state':action[0]
      })

    self.limboActions = []
  

  def train(self, alpha):

    currentState = self.getParams()
    
    ds = [distance(x['state'], currentState) for x in self.stateActions]
    mxs = max([x['score'] for x in self.stateActions])
    ss = [mxs - x['score'] for x in self.stateActions]
    acs = [x['action'] for x in self.stateActions]

    self.setParams(np.add(currentState, np.array(min([[c, a + b] for a, b, c in zip(softmax(*ds), softmax(*ss), acs)], key=lambda x: x[1])[0])*alpha))


  def trim(self, threshold):
    
    newStateActions = []

    for x in self.stateActions:
      if uniform(0, 1) < threshold:
        newStateActions.append(x)

    self.stateActions = newStateActions



from random import randint, uniform

def squish(x, mx, mn):
  return ( x - mn ) / (mx - mn)

def invsquish(x, mx, mn):
  return ( mx - x ) / ( mx - mn )

def distance(x, y):
  return sum(map(lambda a, b: (a - b)**2, x, y))**(1/2)

def normalize(x):
  try:
    return tuple([i / distance((0, 0), x) for i in x])
  except:
    return (0, 0)

class Actor:
  
  def __init__(self):
    raise NotImplementedError("Method hasn't been implemented yet.")

  def act(self):
    raise NotImplementedError("Method hasn't been implemented yet.")

  def getState(self):
    raise NotImplementedError("Method hasn't been implemented yet.")

class Environment:
  
  def __init__(self):
    self.attributions = {}

  def train(self, score, action, state):
    
    if not state in self.attributions:
      
      self.attributions[state] = {"action":action, "score":score}

    elif self.attributions[state]["score"] < score and uniform(0, 1) < (1/3):
      
        self.attributions[state] = {"action":action, "score":score}

  def act(self, current_state):

    self.info = list(map(lambda state: [self.attributions[state]["action"], self.attributions[state]["score"], distance(state, current_state)], self.attributions))

    info = self.info
    
    max_dist = max(info, key=lambda x: x[2])[2]
    min_dist = min(info, key=lambda x: x[2])[2]
    max_score = max(info, key=lambda x: x[1])[1]
    min_score = min(info, key=lambda x: x[1])[1]

    info = list(map(lambda a: [a[0], (invsquish(a[1], max_score, min_score) * 10 + squish(a[2], max_dist, min_dist) * 2) / 12], info))

    return min(info, key=lambda x: x[1])[0]
    
  



import setuptools

with open("README.txt", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hula",
    version="0.0.0.0",
    author="Robert Bassett",
    author_email="robert.bassett.coder@gmail.com",
    description="Experimental learning methods. The first appearance of these methods.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=['numpy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",

    ],
)



from math import e
import numpy as np

def lagzip(l, lagfactor=1):
  for n, x in enumerate(l[:-lagfactor]):
    yield (l[n], l[n+lagfactor])

def softmax(*args):
  return np.exp(args)/np.sum(np.exp(args))

def sigmoid(x):
  return 1/(1+np.exp(-x))

def tanh(x):
  return (np.exp(x)-np.exp(-x)) / (np.exp(x)+np.exp(-x))

def distance(a, b):
  return sum(map(lambda x, y: (x-y)**2, a, b))**(0.5)

def magnitude(a):
  return sum(map(lambda x: x**2, a))**(1/2)

def TwDimMagnitude(a):
  out = []
  for row in a:
    out.append(magnitude(row))
  return magnitude(out)

def ThrDimMagnitude(a):
  out = []
  for matrix in a:
    out.append(TwDimMagnitude(matrix))
  return magnitude(out)

def TwDimDistance(a, b):
  out = []
  for rowa, rowb in zip(a, b):
    out.append(distance(rowa, rowb))
  return magnitude(out)

def ThrDimDistance(a, b):
  out = []
  for matrixa, matrixb in zip(a, b):
    out.append(TwDimDistance(matrixa, matrixb))
  return magnitude(out)

def normalize(a):
  amag = magnitude(a)
  return list(map(lambda x: x/amag, a))
